/*
 Wireless Game Controller - SNES Drop in TX
 Version: 1.0
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 14/03/2019
 Last Modified: 01/06/2019

 */
 
#ifndef cbi
#define cbi(sfr, bit) (_SFR_BYTE(sfr) &= ~_BV(bit))
#endif
#ifndef sbi
#define sbi(sfr, bit) (_SFR_BYTE(sfr) |= _BV(bit))
#endif

#ifndef NULL
#define NULL ((void *)0)
#endif

#define LOW 0
#define HIGH 1
#define false 0
#define true 1

// Watchdog timeouts to values
#define T16MS 0
#define T32MS 1
#define T64MS 2
#define T128MS 3
#define T250MS 4
#define T500MS 5
#define T1S 6
#define T2S 7
#define T4S 8
#define T8S 9

#define LED PD6
#define BUTTON_UP PC0
#define BUTTON_LEFT PC1
#define BUTTON_DOWN PC2
#define BUTTON_RIGHT PC3
#define BUTTON_SELECT PC4
#define BUTTON_START PC5
#define BUTTON_L PC7

#define BUTTON_B PD0
#define BUTTON_Y PD1
#define BUTTON_A PD2
#define BUTTON_X PD3
#define BUTTON_R PD4

#define GB_MODE 1
#define GBA_MODE 2

#define SNES_TX 0x31
#define GC_TX 0x32
#define N64_TX 0x33

// nRF settings (loaded from EEPROM)
uint8_t nrfChannel = 1;
uint8_t nrfTx[5] = "wgbr"; // Last byte loaded from EEPROM
uint8_t nrfPpr = 1;

// EEPROM settings
#define nrfChannelLocation 1
#define nrfTxLocation 2
#define nrfPprLocation 3
#define blankEEPROMByte 255


extern void mirf_init(void);
extern void mirf_config(void);

// Sleep
void system_sleep(void) {
	set_sleep_mode(SLEEP_MODE_PWR_DOWN); // Set sleep mode
	sleep_mode(); // System sleeps here
}

// Setup watchdog timed sequence
void setup_watchdog(uint8_t wdDelay) {
	if (wdDelay > 9) { 
		wdDelay = 9;
	}
	
	uint8_t bb = wdDelay & 7;
	if (wdDelay > 7) { 
		bb |= (1<<5);
	}
	
	MCUSR &= ~(1<<WDRF);
	// Start timed sequence
	WDTCSR |= (1<<WDCE) | (1<<WDE);
	// Set new watchdog timeout value
	WDTCSR = bb | (1<<WDIE);
}

// Turn off the Watchdog
void turn_off_watchdog(void) {
	cli(); // Turn interrupts off
	MCUSR &= ~(1<<WDRF);
	WDTCSR |= (1<<WDCE) | (1<<WDE);
	WDTCSR = 0;
	sei(); // Turn interrupts back on
}

// Watchdog sleep function - setup, sleep and turn off
void watchdog_sleep(uint8_t timeout) { 
	setup_watchdog(timeout);
	system_sleep();
	turn_off_watchdog();
}

// SPI transfer 1 byte and return the result
uint8_t spi_transfer(uint8_t data) {
	SPDR = data;
	
	// Wait for transmission complete
	while (!(SPSR & (1<<SPIF)));
	
	return SPDR;
}

// Write data using SPI
void spi_write_data(uint8_t * dataout, uint8_t len) {
	for (uint8_t i = 0; i < len; i++) {
		spi_transfer(dataout[i]);
	}
}

// Read data using SPI
void spi_read_data(uint8_t * datain, uint8_t len) {
	for (uint8_t i = 0; i < len; i++) {
		datain[i] = spi_transfer(0x00);
	}
}

// Initialise the SPI
void spi_init(void) {
	DDRB |= (1<<PB5); // SPI CLK
	DDRB |= (1<<PB3); // SPI DO
	DDRB &= ~(1<<PB4); // SPI DI
	PORTB |= (1<<PB4); // SPI DI
	DDRB |= (1<<PB2); // SPI SS
	
	// Enable SPI, Master, set clock rate / 4
	SPCR = (1<<SPE)|(1<<MSTR);
}

// Setup
void setup(void) {
	// Setup outputs
	DDRD |= (1<<LED);
	PORTD |= (1<<LED);
	_delay_ms(200);
	PORTD &= ~(1<<LED);
	
	// Load nRF channel and address from EEPROM
	nrfChannel = eeprom_read_byte((uint8_t*) nrfChannelLocation);
	nrfTx[4] = eeprom_read_byte((uint8_t*) nrfTxLocation);
	nrfPpr = eeprom_read_byte((uint8_t*) nrfPprLocation);
	if (nrfChannel == blankEEPROMByte) {
		nrfChannel = 1;
		nrfTx[4] = 1;
		nrfPpr = 1;
	}
	
	// Turn on pull-up resistors on buttons
	PORTC |= (1<<PC7) | (1<<PC5) | (1<<PC4) | (1<<PC3) | (1<<PC2) | (1<<PC1) | (1<<PC0);
	PORTD |= (1<<PD4) | (1<<PD3)|  (1<<PD2) | (1<<PD1) | (1<<PD0);

	// Turn on pull-up resistors on other ports to save power
	PORTB |= (1<<PB7) | (1<<PB6) | (1<<PB0);
	PORTD |= (1<<PD7) | (1<<PD5);
	
	// Turn on pin interrupt on start and select button
	PCMSK1 |= (1<<PCINT13) | (1<<PCINT12);
	PCICR |= (1<<PCIE1);
	
	// mIRF initialise and configure
	spi_init(); // Initialise SPI
	mirf_init();
	_delay_ms(50);
	mirf_config();
	
	sei(); // Turn on interrupts
}
